from datasets import load_dataset
import torch
from torch.utils.data import DataLoader
from einops import rearrange
from transformer_lens import utils


def get_splits(n_vals=32):
    training = load_dataset(
        "NeelNanda/c4-tokenized-2b", split="train", streaming=True
    ).with_format("torch")

    validation = list(training.take(n_vals))
    validation = torch.stack([row["tokens"] for row in validation])

    return training, validation


class ConstrainedAdam(torch.optim.Adam):
    """
    A variant of Adam where some of the parameters are constrained to have unit norm.
    """

    def __init__(self, params, constrained_params, lr):
        super().__init__(params, lr=lr)
        self.constrained_params = list(constrained_params)

    def step(self, closure=None):
        with torch.no_grad():
            for p in self.constrained_params:
                normed_p = p / p.norm(dim=-1, keepdim=True)
                # project away the parallel component of the gradient
                p.grad -= (p.grad * normed_p).sum(dim=-1, keepdim=True) * normed_p
        super().step(closure=closure)
        with torch.no_grad():
            for p in self.constrained_params:
                # renormalize the constrained parameters
                p /= p.norm(dim=-1, keepdim=True)


class Sampler:
    """
    A class for sampling activations from a model at a certain point in the network.
    It stores the activations in a large buffer and returns them in a single tensor.
    """

    def __init__(self, config, dataset, model):
        self.config = config
        self.model = model

        self.d_model = model.cfg.d_model
        self.n_ctx = model.cfg.n_ctx

        assert (
            config.buffer_size % (config.in_batch * self.n_ctx) == 0
        ), "samples must be a multiple of loader batch size"
        self.n_inputs = config.buffer_size // (config.in_batch * self.n_ctx)

        self.loader = DataLoader(dataset, batch_size=config.in_batch)
        self.batches = []

    def collect(self):
        result = rearrange(
            torch.cat(self.batches, dim=0), "... d_model -> (...) d_model"
        )
        self.batches = []
        return result

    def extract(self, batch):
        hook_pt = utils.get_act_name(self.config.point, self.config.layer)
        _, cache = self.model.run_with_cache(
            batch, names_filter=[hook_pt], return_type="loss"
        )
        return cache[hook_pt]

    @torch.inference_mode()
    def __iter__(self):
        self.batches = []

        for batch in self.loader:
            self.batches.append(self.extract(batch["tokens"]))

            if len(self.batches) == self.n_inputs:
                yield self.collect()



import torch
from torch import nn
from einops import einsum
from torch.optim import Adam
from smol_sae.base import BaseSAE, Loss


class RainbowSAE(BaseSAE):
    """
    This is a combination of gated SAE with the Anthropic loss.
    """

    def __init__(self, config, model):
        super().__init__(config, model)
        device = config.device

        W_dec = torch.randn(
            self.n_instances, self.d_hidden, self.d_model, device=device
        )
        W_dec /= torch.norm(W_dec, dim=-2, keepdim=True)
        self.W_dec = nn.Parameter(W_dec)

        self.W_gate = nn.Parameter(W_dec.mT.clone().to(device))
        self.r_mag = nn.Parameter(
            torch.zeros(self.n_instances, self.d_hidden, device=device)
        )

        self.b_gate = nn.Parameter(
            torch.zeros(self.n_instances, self.d_hidden, device=device)
        )
        self.b_mag = nn.Parameter(
            torch.zeros(self.n_instances, self.d_hidden, device=device)
        )
        self.b_dec = nn.Parameter(
            torch.zeros(self.n_instances, self.d_model, device=device)
        )

        self.relu = nn.ReLU()
        self.optimizer = Adam(self.parameters(), lr=config.lr, betas=(0.9, 0.999))

    def encode(self, x):
        gated_preact = (
            einsum(self.W_gate, x - self.b_dec, "inst d h, ... inst d -> ... inst h")
            + self.b_gate
        )
        gate = (gated_preact > 0).float()

        W_mag = einsum(
            torch.exp(self.r_mag), self.W_gate, "inst h, inst d h -> inst d h"
        )
        magnitude = (
            einsum(W_mag, x - self.b_dec, "inst d h, ... inst d -> ... inst h")
            + self.b_mag
        )
        mag_act = torch.relu(magnitude)

        hidden_act = mag_act * gate
        gated_act = torch.relu(gated_preact)

        return hidden_act, gated_act

    def decode(self, h):
        return einsum(h, self.W_dec, "... inst h, inst h d -> ... inst d") + self.b_dec

    def loss(self, x, x_hid, x_hat, fraction, gated_act):
        recons_losses = ((x - x_hat) ** 2).mean(dim=0).sum(dim=-1)

        W_dec_clone = self.W_dec.clone().detach()
        b_dec_clone = self.b_dec.clone().detach()

        norm = W_dec_clone.norm(dim=-1)
        lambda_ = min(1, fraction * 20)
        sparsity_losses = lambda_ * einsum(
            gated_act, norm, "batch inst hidden, inst hidden -> inst batch"
        ).mean(dim=-1)

        gated_recons = (
            einsum(
                gated_act,
                W_dec_clone,
                "batch inst hidden, inst hidden d -> batch inst d",
            )
            + b_dec_clone
        )
        aux_losses = ((x - gated_recons) ** 2).mean(dim=0).sum(dim=-1)

        return Loss(recons_losses, sparsity_losses, aux_losses)

    def calculate_metrics(self, x_hid, losses, *args):
        metrics = super().calculate_metrics(x_hid, losses, *args)
        metrics |= {
            f"auxiliary_loss/{i}": loss for i, loss in enumerate(losses.auxiliary)
        }
        return metrics



import torch
from torch import nn
from einops import einsum
from smol_sae.base import BaseSAE, Loss
from smol_sae.utils import ConstrainedAdam


class VanillaSAE(BaseSAE):
    """
    An ordinary SAE, no ghost grads or resampling.
    """

    def __init__(self, config, model):
        super().__init__(config, model)
        device = config.device

        W_dec = torch.randn(
            self.n_instances, self.d_hidden, self.d_model, device=device
        )
        W_dec /= torch.norm(W_dec, dim=-2, keepdim=True)
        self.W_dec = nn.Parameter(W_dec)

        W_enc = W_dec.mT.clone().to(device)
        self.W_enc = nn.Parameter(W_enc)

        self.b_enc = nn.Parameter(
            torch.zeros(self.n_instances, self.d_hidden, device=device)
        )
        self.b_dec = nn.Parameter(
            torch.zeros(self.n_instances, self.d_model, device=device)
        )

        self.relu = nn.ReLU()
        self.optimizer = ConstrainedAdam(
            self.parameters(), [self.W_dec], lr=self.config.lr
        )

    def encode(self, x):
        return (
            self.relu(
                einsum(
                    x - self.b_dec,
                    self.W_enc,
                    "... inst d, inst d hidden -> ... inst hidden",
                )
                + self.b_enc
            ),
        )

    def decode(self, h):
        return (
            einsum(h, self.W_dec, "... inst hidden, inst hidden d -> ... inst d")
            + self.b_dec
        )

    def loss(self, x, x_hid, x_hat, fraction):
        reconstruction = ((x_hat - x) ** 2).mean(0).sum(dim=-1)

        lambda_ = min(1, fraction * 20)
        sparsity = x_hid.abs().mean(dim=0).sum(-1)

        return Loss(
            reconstruction,
            lambda_ * sparsity,
            torch.zeros(self.n_instances, device=x.device),
        )



from transformer_lens import HookedTransformer
import torch

from smol_sae.base import Config
from utils import get_splits, Sampler
from smol_sae.rainbow import RainbowSAE

if __name__ == "__main__":
    torch.backends.cudnn.benchmark = True
    model = HookedTransformer.from_pretrained("gelu-1l").cuda()
    train, validation = get_splits()
    # This equates to about 65M tokens
    # This takes about 16 GB of GPU memory
    config = Config(
        n_buffers=500,
        expansion=4,
        buffer_size=2**17,
        sparsities=(0.01, 0.02, 0.04, 0.07, 0.14, 0.27, 0.52, 1.00),
    )
    sampler = Sampler(config, train, model)
    sae = RainbowSAE(config, model).cuda()

    sae.train(sampler, model, validation, log=True)



import torch
from torch import nn
from einops import einsum
from torch.optim import Adam
from smol_sae.base import BaseSAE, Loss


class AnthropicSAE(BaseSAE):
    """
    The main difference between this is the loss function.
    Specifically, it uses the activation * the output norm as the sparsity term.
    https://transformer-circuits.pub/2024/april-update/index.html#training-saes
    """

    def __init__(self, config, model):
        super().__init__(config, model)
        device = config.device

        W_dec = torch.randn(
            self.n_instances, self.d_hidden, self.d_model, device=device
        )
        W_dec /= torch.norm(W_dec, dim=-2, keepdim=True)
        self.W_dec = nn.Parameter(W_dec)

        W_enc = W_dec.mT.clone().to(device)
        self.W_enc = nn.Parameter(W_enc)

        # Contrary to Anthropic, we actually still use a decoder norm because it seems more logical.
        self.b_enc = nn.Parameter(
            torch.zeros(self.n_instances, self.d_hidden, device=device)
        )
        self.b_dec = nn.Parameter(
            torch.zeros(self.n_instances, self.d_model, device=device)
        )

        self.relu = nn.ReLU()
        self.optimizer = Adam(self.parameters(), lr=self.config.lr, betas=(0.9, 0.999))

    def encode(self, x):
        return (
            self.relu(
                einsum(
                    x - self.b_dec,
                    self.W_enc,
                    "... inst d, inst d hidden -> ... inst hidden",
                )
                + self.b_enc
            ),
        )

    def decode(self, h):
        return (
            einsum(h, self.W_dec, "... inst hidden, inst hidden d -> ... inst d")
            + self.b_dec
        )

    def loss(self, x, x_hid, x_hat, fraction):
        reconstruction = ((x_hat - x) ** 2).mean(0).sum(dim=-1)

        norm = self.W_dec.norm(dim=-1)
        lambda_ = min(1, fraction * 20)

        sparsity = einsum(
            x_hid, norm, "batch inst hidden, inst hidden -> batch inst"
        ).mean(dim=0)

        return Loss(
            reconstruction,
            lambda_ * sparsity,
            torch.zeros(self.n_instances, device=x.device),
        )



from torch import nn
import torch
from functools import partial
from transformer_lens import utils
from dataclasses import dataclass
from einops import repeat
from torch.utils.data import DataLoader
from torch.optim.lr_scheduler import LambdaLR
import wandb
from tqdm import tqdm
from collections import namedtuple

Loss = namedtuple("Loss", ["reconstruction", "sparsity", "auxiliary"])


@dataclass
class Config:
    buffer_size: int = 2**18  # ~250k tokens
    n_buffers: int = 100  # ~25M tokens

    # transformer_lens specific
    point: str = "resid_mid"
    layer: int = 0

    in_batch: int = 32
    out_batch: int = 4096

    expansion: int = 4
    lr: float = 1e-4

    validation_interval: int = 1000

    sparsities: tuple = (0.01, 0.1, 1)  # L1 loss coefficients
    device: str = "cuda"

    @property
    def n_instances(self):
        return len(self.sparsities)


class BaseSAE(nn.Module):
    """
    Base class for all Sparse Auto Encoders.
    Provides a common interface for training and evaluation.
    """

    def __init__(self, config, model) -> None:
        super().__init__()
        self.config = config

        self.d_model = model.cfg.d_model
        self.n_ctx = model.cfg.n_ctx
        self.d_hidden = self.config.expansion * self.d_model
        self.n_instances = len(config.sparsities)

        self.steps_not_active = torch.zeros(self.n_instances, self.d_hidden)
        self.sparsities = torch.tensor(config.sparsities).to(config.device)
        self.step = 0

    def decode(self, h):
        return h

    def encode(self, x):
        return x

    def forward(self, x):
        x_hid, *_ = self.encode(x)
        return self.decode(x_hid)

    def loss(self, x, x_hid, x_hat, fraction, *args):
        pass

    @classmethod
    def from_pretrained(cls, path, device="cpu", *args):
        state = torch.load(path)
        new = cls(*args)
        return new.load_state_dict(state).to(device)

    def calculate_metrics(self, x_hid, losses, *args):
        activeness = x_hid.sum(0)
        self.steps_not_active[activeness > 0] = 0

        metrics = dict(step=self.step)

        for i in range(self.n_instances):
            metrics[f"dead_fraction/{i}"] = (
                (self.steps_not_active[i] > 2).float().mean().item()
            )

            metrics[f"reconstruction_loss/{i}"] = losses.reconstruction[i].item()
            metrics[f"sparsity_loss/{i}"] = losses.sparsity[i].item()
            metrics[f"auxiliary_loss/{i}"] = losses.auxiliary[i].item()

            metrics[f"l1/{i}"] = x_hid[..., i, :].sum(-1).mean().item()
            metrics[f"l0/{i}"] = (x_hid[..., i, :] > 0).float().sum(-1).mean().item()

        self.steps_not_active += 1
        self.step += 1

        return metrics

    def train(self, sampler, model, validation, log=True):
        if log:
            wandb.init(project="sae")

        self.step = 0
        steps = self.config.n_buffers * (
            self.config.buffer_size // self.config.out_batch
        )

        scheduler = LambdaLR(
            self.optimizer, lr_lambda=lambda t: min(5 * (1 - t / steps), 1.0)
        )

        for buffer, _ in tqdm(
            zip(sampler, range(self.config.n_buffers)), total=self.config.n_buffers
        ):
            loader = DataLoader(
                buffer, batch_size=self.config.out_batch, shuffle=True, drop_last=True
            )
            for x in loader:
                x = repeat(x, "... d -> ... inst d", inst=self.n_instances)
                x_hid, *rest = self.encode(x)
                x_hat = self.decode(x_hid)

                losses = self.loss(x, x_hid, x_hat, self.step / steps, *rest)
                metrics = self.calculate_metrics(x_hid, losses, self.step, *rest)

                loss = (
                    losses.reconstruction
                    + self.sparsities * losses.sparsity
                    + losses.auxiliary
                ).sum()

                self.optimizer.zero_grad()
                loss.backward()
                self.optimizer.step()
                scheduler.step()

                if log and (self.step % self.config.validation_interval == 0):
                    clean_loss, losses = self.patch_loss(model, validation)
                    metrics |= {
                        f"patch_loss/{i}": (loss.item() - clean_loss) / clean_loss
                        for i, loss in enumerate(losses)
                    }

                if log:
                    wandb.log(metrics)

        if log:
            wandb.finish()

    @torch.inference_mode()
    def patch_loss(self, model, validation):
        losses = torch.zeros(self.n_instances, device=self.config.device)
        hook_pt = utils.get_act_name(self.config.point, self.config.layer)

        baseline, cache = model.run_with_cache(
            validation, return_type="loss", names_filter=[hook_pt]
        )

        x = repeat(cache[hook_pt], "... d -> ... inst d", inst=self.n_instances)
        x_hat = self.forward(x)

        def patch_hook(act, hook, inst_id):
            return x_hat[:, :, inst_id]

        # run model with recons patched in per instance
        for inst_id in range(self.n_instances):
            fwd_hook = partial(patch_hook, inst_id=inst_id)
            loss = model.run_with_hooks(
                validation, return_type="loss", fwd_hooks=[(hook_pt, fwd_hook)]
            )
            losses[inst_id] = loss.item()

        return baseline, losses



import torch
import pytest
from smol_sae.base import Config
from smol_sae.vanilla import VanillaSAE
from smol_sae.utils import get_splits
from transformer_lens import HookedTransformer


@pytest.fixture
def device():
    # if torch.cuda.is_available():
    #     return "cuda"
    # elif torch.backends.mps.is_available():
    #     return "mps"
    # else:
    return "cpu"


@pytest.fixture
def model(device):
    return HookedTransformer.from_pretrained("gelu-1l").to(device)


@pytest.fixture
def config(device):
    return Config(
        n_buffers=2,
        expansion=2,
        in_batch=32,
        out_batch=32,
        buffer_size=2**8,
        sparsities=(0.1, 1.0),
        device=device,
    )


@pytest.fixture()
def sae(config, model):
    return VanillaSAE(config, model)


def test_vanilla_sae_forward(model, device, sae):
    n_batch = 2
    # test forward
    input = torch.randn(n_batch, sae.n_instances, model.cfg.d_model).to(device)
    output = sae(input)
    assert output.shape == input.shape


def test_vanilla_sae_patch_loss(model, device, sae: VanillaSAE):
    _, validation = get_splits()
    print(validation.shape)
    sae.patch_loss(model, validation)



import torch
from smol_sae.base import Config
from smol_sae.vanilla import VanillaSAE
from transformer_lens import HookedTransformer

device = "cuda" if torch.cuda.is_available() else "cpu"
config = Config(
    n_buffers=100, expansion=4, buffer_size=2**8, sparsities=(0.1, 1.0), device=device
)

model = HookedTransformer.from_pretrained("gelu-1l").to(device)
sae = VanillaSAE(config, model)
n_batch = 32
sae_input = torch.randn(n_batch, sae.n_instances, model.cfg.d_model).to(device)
sae_output = sae(sae_input)
